package jp.ac.nagoya_cu.sda.paper.behavior;

import jp.ac.nagoya_cu.sda.paper.*;
import jp.ac.nagoya_cu.sda.paper.interpolator.RotationViewInterpolator;
import jp.ac.nagoya_cu.sda.paper.shape.LineShape;
import jp.ac.nagoya_cu.sda.paper.shape.OrigamiShape;
import jp.ac.nagoya_cu.sda.paper.shape.Plane;
import jp.ac.nagoya_cu.sda.paper.shape.PointShape;

import com.sun.j3d.utils.picking.*;
import java.awt.event.*;
import java.util.*;

import javax.media.j3d.*;
import javax.vecmath.*;

import sun.security.krb5.internal.p;

public class FoldingBehavior extends MouseBehavior {

	private PickCanvas pickCanvas;
	private Canvas3D canvas;
	private Paper paper;
	private boolean picked = false;
	private TransformGroup roottrans=null;
	private boolean isRotateRoot=false;

	// sbLOplanei[
	private Plane plane = null;
	private FoldingLine axisLine = null;
	private Point3d pickedPoint=new Point3d();

//	int xpos = 0, ypos = 0;
	int oldx = 0, oldy = 0;

	private Shape3D shpFoldingLine=new Shape3D();
	private Shape3D shpLongFoldingLine=new Shape3D();
	private Shape3D shpSelectedPlaneR =new Shape3D();
	private Shape3D shpSelectedPlaneB =new Shape3D();
	private Shape3D shpCircle=new Shape3D();
	
	private RotationViewInterpolator rotateInterpolator=null;

	public FoldingBehavior(Canvas3D canvas, Paper paper,TransformGroup roottrans, Bounds bounds) {
		pickCanvas = new PickCanvas(canvas, paper);
		pickCanvas.setMode(PickCanvas.GEOMETRY_INTERSECT_INFO);
		pickCanvas.setTolerance(2f);
		setSchedulingBounds(bounds);
		this.paper = paper;
		this.canvas=canvas;
		this.roottrans=roottrans;
		
		addShapeToPaper(shpFoldingLine,OriUtil.createLineAppearance(new Color3f(0,1,1),4,LineAttributes.PATTERN_SOLID));
		addShapeToPaper(shpLongFoldingLine,OriUtil.createLineAppearance(new Color3f(1,.5f,0),1,LineAttributes.PATTERN_SOLID));
		addShapeToPaper(shpSelectedPlaneR,OriUtil.createSelectedPlaneAppearance(new Color3f(0,1,0),true));
		addShapeToPaper(shpSelectedPlaneB,OriUtil.createSelectedPlaneAppearance(new Color3f(0,1,0),false));
		addShapeToPaper(shpCircle,createPlaneAppearance(new Color3f(.5f,.5f,1f),0.5f));
		
		rotateInterpolator=new RotationViewInterpolator();
		rotateInterpolator.setSchedulingBounds(bounds);
		rotateInterpolator.setTarget(roottrans);
		rotateInterpolator.setMinimumAngle(0);
		rotateInterpolator.setMaximumAngle((float)Math.PI /4f);
		paper.addChild(rotateInterpolator);

	}


	
	public void onMousePressed(MouseEvent mevent) {

		oldx = mevent.getX();
		oldy = mevent.getY();
		boolean flag=false;
		if (getMouseButtonStatus(mevent)==MouseEvent.BUTTON1_DOWN_MASK) {

			plane=pickPlane(mevent,pickedPoint);
			
			if (plane !=null) {
				showPlane(plane);
				axisLine = getFoldingLine(plane,pickedPoint); // 
				if(axisLine!=null){
					picked = true;
					flag=true;
					showFoldingLine(axisLine);
					showLongFoldingLine(axisLine);
					showFoldingCircle(axisLine,pickedPoint);
					if(isVertical(axisLine)){
						System.out.println("");
						
						Alpha alpha=new Alpha(1,100);
						alpha.setStartTime(System.currentTimeMillis());
						alpha.setMode(Alpha.INCREASING_ENABLE);
						
						rotateInterpolator.setAlpha(alpha);
						rotateInterpolator.setTransformAxis(createAxisTransform(axisLine,pickedPoint));
						isRotateRoot=true;
					}else{
						isRotateRoot=false;
					}
				}
			}
		}
		if(!flag){
			picked=false;
			showReset();
		}

	}
	public void onMouseReleased(MouseEvent mevent){
		if (picked) {
			if(isRotateRoot){
				Alpha alpha=new Alpha(1,0);
				alpha.setStartTime(System.currentTimeMillis());
				alpha.setMode(Alpha.DECREASING_ENABLE);
				alpha.setDecreasingAlphaDuration(100);
				rotateInterpolator.setAlpha(alpha);
			}

			picked = false;
			paper.unlockAllVertices();
			paper.pushHistory();
		}
	}
	public void onMouseDragged(MouseEvent mevent){
		
		int xpos=mevent.getX();
		int ypos=mevent.getY();
		
		if (picked && axisLine != null) {
			translate(xpos, ypos, oldx, oldy);
			oldx = xpos;
			oldy = ypos;
			showPlane(plane);
			showLongFoldingLine(axisLine);
			showFoldingLine(axisLine);
			showFoldingCircle(axisLine,pickedPoint);
		}else{
			showReset();
		}
	}
	public void onMouseMoved(MouseEvent mevent) {
		Point3d crrPt=new Point3d();
		Plane plane=pickPlane(mevent,crrPt);

		boolean flag=false;
		if (plane !=null) {
			showPlane(plane);
			FoldingLine axisLine = getFoldingLine(plane,crrPt); // 
			if(axisLine!=null){
				flag=true;
				showLongFoldingLine(axisLine);
				showFoldingLine(axisLine);
				showFoldingCircle(axisLine,crrPt);
			}
		}
		if(!flag){
			showReset();
		}

	}
	
	
	
	private Vector3d vecZ=new Vector3d(0,0,-1);
	Transform3D isvt3dLV=new Transform3D();
	/**
	 * linesɑ΂Đǂ
	 */
	private boolean isVertical(Line line){
		Vector3d v=line.createDirection();
		paper.getLocalToVworld(isvt3dLV);
		isvt3dLV.transform(v);
		if(Math.abs(90-vecZ.angle(v)/Math.PI*180) < 15d){
			return true;
		}else{
			return false;
		}
	}
	
	
	private void createInterpolator(Line line,Point3d crrPt){
		Alpha alpha=new Alpha();
		alpha.setStartTime(System.currentTimeMillis());
		alpha.setLoopCount(1);
		alpha.setIncreasingAlphaDuration(100);
		
		Transform3D t3dNew=createAxisTransform(line,crrPt);
	}
	
	private Transform3D createAxisTransform(Line line,Point3d crrPt){
		Transform3D t3dVL=new Transform3D();
		paper.getLocalToVworld(t3dVL);
		t3dVL.invert();

		Vector3d vecY=new Vector3d(0,1,0);
		Vector3d vecCrrPt=new Vector3d(crrPt);
		t3dVL.transform(vecY);
		
		Vector3d vecCenter=new Vector3d(line.closestPointOnLine(crrPt));
		vecCrrPt.sub(vecCenter);

		t3dVL.invert();
		double arg=vecY.angle(vecCrrPt);
		Vector3d axisVW=new Vector3d();
		axisVW.cross(vecY,vecCrrPt);
		t3dVL.transform(axisVW);
		
		AxisAngle4d aa=new AxisAngle4d(axisVW,arg);
		Transform3D t3dNew=new Transform3D();
		t3dNew.set(aa);
		t3dVL.transform(vecCenter);
		t3dNew.setTranslation(vecCenter);
		return t3dNew;
	}
	
	
	
	/**
	 * NbNIuWFNgԂ
	 * @param mevent MouseEvent
	 * @param crrPt IuWFNgŁA}EXŃNbNʒuɑW𒲂ׂĊi[
	 * @return NbNꂽIuWFNgBnull
	 */
	private OrigamiShape pickOrigamiShape(MouseEvent mevent,Point3d crrPt) {

		// PickCanvas ăsbLOĂm[hT
		pickCanvas.setShapeLocation(mevent);
		// LineArray  Vertex ̗dȂĂꍇ
		// Vertex o߂ɁApickAll() őSĎo
		PickResult[] pr = pickCanvas.pickAllSorted();

		if (pr != null) {
			crrPt.set(OriUtil.getPickedLocation(canvas,mevent.getPoint(),pr));
			return OriUtil.getPickedOrigamiShape(pr);
		}
		return null;
	}
	
	private Plane pickPlane(MouseEvent mevent,Point3d crrPt){
		pickCanvas.setShapeLocation(mevent);
		PickResult[] res=pickCanvas.pickAllSorted();
		
		if(res!=null){
			for(PickResult pr:res){
				if(pr.getNode(PickResult.BRANCH_GROUP) instanceof Plane){
					try{
						crrPt.set(pr.getIntersection(0).getPointCoordinates());
					}catch(RuntimeException e){
						System.out.println("Interp point outside triangle");
						crrPt.set(pr.getIntersection(0).getClosestVertexCoordinates());
					}
					return (Plane)pr.getNode(PickResult.BRANCH_GROUP);
				}
			}
		}
		return null;
	}

	// ۂɖʂ𓮂
	private void translate(int xpos, int ypos, int oldx, int oldy) {
		
		//}EXW3ɕϊ
		Point3d eyePos = OriUtil.getCenterEyePosVW(canvas);
		Point3d _p1=OriUtil.getPointVW(canvas,oldx,oldy);
		Point3d _p2=OriUtil.getPointVW(canvas,xpos,ypos);

		//[JWn
		Transform3D t3dVL=new Transform3D();
		paper.getLocalToVworld(t3dVL);
		t3dVL.invert();
		t3dVL.transform(eyePos);
		t3dVL.transform(_p1);
		t3dVL.transform(_p2);

		//NbN_ʂ莲ƐȖʂݒ肵A̖ʏł̃}EẌʒu擾
		Point3d p=getCrossPoint(axisLine.getPointA(),axisLine.getPointB());	//ʂƎƂ̌_
		Point3d p1=getCrossPoint(eyePos,_p1);
		Point3d p2=getCrossPoint(eyePos,_p2);

		//pxvZ
		Vector3d v1=new Vector3d();
		Vector3d v2=new Vector3d();
		v1.sub(p1,p);
		v2.sub(p2,p);
		double angle=v1.angle(v2);

		//px̕\␳
		Vector3d norm=new Vector3d();
		norm.cross(v1,v2);
		Vector3d rightside=new Vector3d();
		rightside.sub(p,eyePos);
		if(rightside.dot(axisLine.createDirection())>0){
			rightside.scale(-1);
		}
		if(norm.dot(rightside)>0){
			angle *= -1;
		}
		
		Matrix4d maxis = getMatrix(axisLine,angle);

		Plane currPlane = plane;
		for(VertexPoint vp:currPlane.getPlaneGeometry().getAllVertexPoints()){
			maxis.transform(vp);
			//ړ̖ʂ͌ŒtO𗧂Ă
			vp.lock();
		}

		// ̖ʂŒ肷
		Plane anotherPlane = paper.getAnotherPlane(axisLine,currPlane);
		for(VertexPoint vp:anotherPlane.getPlaneGeometry().getAllVertexPoints()){
			vp.lock();
		}
	}

	// Ɖ]p]s𐶐
	protected Matrix4d getMatrix(FoldingLine axis, double arg) {
		// ]s񐶐
		AxisAngle4d axisAngle = new AxisAngle4d();
		axisAngle.x = axis.getPointB().x - axis.getPointA().x;
		axisAngle.y = axis.getPointB().y - axis.getPointA().y;
		axisAngle.z = axis.getPointB().z - axis.getPointA().z;
		axisAngle.angle = arg;

		// ]sMatrix4d^ɕϊ
		Matrix4d maxis = new Matrix4d();
		maxis.set(axisAngle);

		// ̍W擾ăZbg
		Vector3d v1 = new Vector3d(axis.getPos());
		Vector3d v2 = new Vector3d(v1);

		// v2]
		maxis.transform(v2);
		// System.out.println("vfrom "+v1 + " vto "+v2);

		// ]Ăꂽǂ
		v1.sub(v2);
		maxis.setTranslation(v1);

		return maxis;
	}

	private double dot(Tuple3d v1, Tuple3d v2){
		return v1.x*v2.x+v1.y*v2.y+v1.z*v2.z;
	}
	
	protected Point3d getCrossPoint(Point3d pstart,Point3d pend){
		Vector3d normal=axisLine.createDirection();
		normal.normalize();

		Point3d p=pickedPoint;
		
		Vector3d vecD = new Vector3d(); // xNg
		vecD.sub(pend, pstart);

		// _߂
		double distance=dot(normal,p);
		double t = (distance - dot(normal,pstart))/ normal.dot(vecD);
		Point3d crossPoint = new Point3d();
		crossPoint.scaleAdd(t, vecD, pstart);
		return crossPoint;
	}
	
	private void showPlane(Plane plane){
		GeometryArray g=(GeometryArray)plane.getPlaneGeometry().getGeometry();
		shpSelectedPlaneR.setGeometry(g);
		shpSelectedPlaneB.setGeometry(g);
	}

	/**
	 * w肳ꂽLinenCCg
	 */
	private void showFoldingLine(Line line){
		Point3d p1=new Point3d(line.getPointA());
		Point3d p2=new Point3d(line.getPointB());
		showLine(shpFoldingLine,p1,p2);
	}
	private void showLongFoldingLine(Line line){
		Point3d p1=new Point3d();
		p1.sub(line.getPointB(),line.getPointA());
		p1.scaleAdd(50,p1,line.getPointA());
		Point3d p2=new Point3d();
		p2.sub(line.getPointA(),p1);
		p2.scaleAdd(2,p2,p1);
		
		showLine(shpLongFoldingLine,p1,p2);
	}
	/**
	 * w肳ꂽ_Ɠ_OłȂ
	 * @param lshape
	 * @param p0
	 * @param p1
	 */
	private void showLine(Shape3D lshape,Point3d p0,Point3d p1){
		LineArray g=new LineArray(2,LineArray.COORDINATES);
		g.setCoordinate(0,p0);
		g.setCoordinate(1,p1);
		lshape.setGeometry(g);
	}
	
	private void showFoldingCircle(FoldingLine axis,Point3d p){

		Point3d center=axis.closestPointOnLine(p);

		int plen=24;
		int[] cnt={plen+2};
		TriangleFanArray g=new TriangleFanArray(plen+2,LineStripArray.COORDINATES,cnt);
		g.setCoordinate(0,center);
		
		Matrix4d mat=new Matrix4d();
		mat.set(new AxisAngle4d(axis.createDirection(),Math.PI*2d/plen));
		
		Point3d pt=new Point3d();
		Point3d pt2=new Point3d();
		pt.sub(p,center);
		for(int i=1;i<plen+2;i++){
			mat.transform(pt);
			pt2.add(pt,center);
			g.setCoordinate(i,pt2);
		}
		shpCircle.setGeometry(g);
	}
	
	private void showReset(){
		shpSelectedPlaneR.setGeometry(null);
		shpSelectedPlaneB.setGeometry(null);
		shpFoldingLine.setGeometry(null);
		shpLongFoldingLine.setGeometry(null);
		shpCircle.setGeometry(null);
	}
	
	private void addShapeToPaper(Shape3D shp,Appearance ap){
		shp.setCapability(Shape3D.ALLOW_GEOMETRY_WRITE);
		shp.setCapability(Shape3D.ALLOW_APPEARANCE_WRITE);
		shp.setPickable(false);
		shp.setAppearance(ap);
		this.paper.addChild(shp);
	}
	private static Appearance createPlaneAppearance(Color3f color,float opacity) {
		ColoringAttributes ca=new ColoringAttributes(color,ColoringAttributes.NICEST);
		
		PolygonAttributes pa = new PolygonAttributes();
		pa.setPolygonMode(PolygonAttributes.POLYGON_FILL);
		pa.setBackFaceNormalFlip(true);
		pa.setCullFace(PolygonAttributes.CULL_NONE);
		
		TransparencyAttributes ta=new TransparencyAttributes(TransparencyAttributes.NICEST,opacity);

		Appearance ap = new Appearance();
		ap.setPolygonAttributes(pa);
		ap.setTransparencyAttributes(ta);
		ap.setColoringAttributes(ca);

		return ap;
	}
	
	private Point3d flp=new Point3d();
	private Point3d flpe=new Point3d();
	private Triangle fltr=new Triangle();
	private Transform3D flt3dvl=new Transform3D();
	private FoldingLine getFoldingLine(Plane plane,Point3d crrPt){
		PlaneGeometry pg=plane.getPlaneGeometry();
		flp.set(0,0,0);
		for(VertexPoint vp:pg.getAllVertexPoints()){
			flp.add(vp);
		}
		flp.scale(1d/pg.getAllVertexPoints().length);
		
		flpe.set(OriUtil.getCenterEyePosVW(canvas));
		paper.getLocalToVworld(flt3dvl);
		flt3dvl.invert();
		flt3dvl.transform(flpe);
		
		fltr.set(flpe,flp,crrPt);
		double dist=Double.MAX_VALUE;
		FoldingLine ln=null;
		for(Line line:pg.getAllLine()){
			if(! (line instanceof FoldingLine)){
				continue;
			}
			//ʂђʂĂȂ΃XLbv
			if(fltr.isRightSide(line.getPointA()) == fltr.isRightSide(line.getPointB())){
				continue;
			}
			Point3d pc=fltr.intersect(line.createStartPoint(),line.createDirection());
			double dist1=pc.distanceSquared(flp);
			double dist2=pc.distanceSquared(crrPt);
			if(dist1>dist2){
				continue;
			}
			if(dist1 < dist){
				dist=dist1;
				ln=(FoldingLine)line;
			}
		}
		if(ln!=null){
			return (FoldingLine)ln;
		}else{
			return null;
		}
	}
	
}